import java.lang.Object;
import java.util.Enumeration;
import ListException;

// We use two auxilary classes: ListHeader and ListElement.
// An object of class ListHeader contains only the pointers
// to the next and previous ListHeaders. This class is used
// to link elements in the list. Also an objects of this class
// is used to store the header of a list.
//
//     ListHeader:
//          +------+
//          | next |
//          | prev |
//          +------+
//
class ListHeader extends Object {
    ListHeader next;
    ListHeader prev;

    ListHeader() {
        next = null;
        prev = null;
    }

    void link(ListHeader h) {
        next = h;
        h.prev = this;
    }

    void release() { next = null; prev = null; }
}

// Class ListElement is derived from ListHeader.
// It inherits all members and methods of ListHeader. The only
// additional member is the value of list element.
// The reason for creating of two separate classes is that we do not
// want to waste a memory for "element" member in a head of
// a list. A list head has the type "ListHeader", all other
// blocks have the type "ListElement".
//     ListElement:
//          +---------+
//          | next    | ListHeader
//          | prev    |
//          |- - - - -|
//          | element |
//          +---------+
//
class ListElement extends ListHeader {
    Object element;

    ListElement() {
        super();
        element = null;
    }

    void release() { element = null; super.release(); }
}

// The idea of implementation of L2List:
// All elements of the list are linked into ring.
// The ring begins with headList element ("header"). If list is
// empty, then the header is linked with itself.
//
//     headList       First elem.     Second elem ... Last element
//     +------+        +------+        +------+        +------+
// +-->|      | ------>|      | ------>|      |     -->|      |---->+
// |   |      |        |      |        |      | . . .  |      |     |
// |+<-|      | < - - -|      | < - - -|      |   < - -|      |<- -+|
// ||  +------+        |      |        |      |        |      |    ||
// ||                  +------+        +------+        +------+    ||
// |+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - --+|
// +----------------------------------------------------------------+
//
// The next element for the headList is the first element of list,
// the previous -- the last element of list.
//

public class L2List {
    private ListHeader headList; // Head of the ring of list elements
    private ListHeader pointer;  // Pointer to element AFTER pointer (element
                                 //     BEFORE can be obtained via m_Prev)
    private int numElements;     // Number of elements in the list
    private int pointerPos;      // Number of elements before a pointer

    public L2List() {
        headList = new ListHeader(); // Create the header of list.
        headList.link(headList);     // Link the header with itself.
        pointer = headList;
        numElements = 0;
        pointerPos = 0;
    }

    public void removeAll() {
	moveToBeg();
	while (!inEnd()) {
	    removeAfter();
	}
    }

    public void moveToBeg() {
        pointer = headList.next;
        pointerPos = 0;
    }

    public void moveToEnd() {
        pointer = headList;
        pointerPos = size();
    }

    public boolean inBeg() {
        return (pointer.prev == headList);
    }

    public boolean inEnd() {
        return (pointer == headList);
    }

    public void moveForward() throws ListException {
        if (inEnd()) {
            throw new ListException("moveForward: End of list");
        } else {
            pointer = pointer.next;
            pointerPos++;
        }
    }

    public void moveBack() throws ListException {
        if (inBeg()) {
            throw new ListException("moveBack: Beginning of list");
        } else {
            pointer = pointer.prev;
            pointerPos--;
        }
    }

    /**
     * Element before a pointer.
     */
    public Object elementBefore() throws ListException {
        if (inBeg()) {
            throw new ListException("elementBefore: Beginning of list");
        } else {
            return ((ListElement) pointer.prev).element;
        }
    }

    /**
     * Set element before a pointer. Return a previous value of element.
     */
    public Object setElementBefore(Object e) throws ListException {
        if (inBeg()) {
            throw new ListException("setElementBefore: Beginning of list");
        } else {
            Object old = ((ListElement) pointer.prev).element;
            ((ListElement) pointer.prev).element = e;
            return old;
        }
    }

    /**
     * Set element after a pointer. Return a previous value of element.
     */
    public Object setElementAfter(Object e) throws ListException {
        if (inEnd()) {
            throw new ListException("setElementAfter: End of list");
        } else {
            Object old = ((ListElement) pointer).element;
            ((ListElement) pointer).element = e;
            return old;
        }
    }

    /**
     * Element after a pointer.
     */
    public Object elementAfter() throws ListException {
        if (inEnd()) {
            throw new ListException("elementAfter: End of list");
        } else {
            return ((ListElement) pointer).element;
        }
    }

    /**
     * Add element before a pointer.
     */
    public void addBefore(Object e) {
        ListElement block = new ListElement();
        block.element = e;
        pointer.prev.link(block);
        block.link(pointer);
        numElements++;
        pointerPos++;
    }

    /**
     * Remove element before a pointer.
     */
    public void removeBefore() throws ListException {
        if (inBeg()) {
            throw new ListException("removeBefore: Beginning of list");
        } else {
            ListHeader block = pointer.prev;
            block.prev.link(pointer);
            block.release();
            numElements--;
            pointerPos--;
        }
    }

    /**
     * Add element after a pointer.
     */
    public void addAfter(Object e) {
        ListElement block = new ListElement();
        block.element = e;
        pointer.prev.link(block);
        block.link(pointer);
        pointer = block;
        numElements++;
    }

    /**
     * Remove element after a pointer.
     */
    public void removeAfter() throws ListException {
        if (inEnd()) {
            throw new ListException("removeAfter: End of list");
        } else {
            ListHeader block = pointer;
            pointer = block.next;
            block.prev.link(pointer);
            block.release();
            numElements--;
        }
    }

    public int size() { return numElements; }

    class ListEnumeration implements Enumeration {
        ListHeader currentBlock;

        public ListEnumeration(ListHeader startBlock) {
            currentBlock = startBlock;
        }

        public boolean hasMoreElements() {
            return (currentBlock != headList);
        }

        public Object nextElement() {
            Object res = ((ListElement) currentBlock).element;
            currentBlock = currentBlock.next;
            return res;
        }
    }

    class ListReverseEnumeration implements Enumeration {
        ListHeader currentBlock;

        public ListReverseEnumeration(ListHeader startBlock) {
            currentBlock = startBlock;
        }

        public boolean hasMoreElements() {
            return (currentBlock != headList);
        }

        public Object nextElement() {
            Object res = ((ListElement) currentBlock).element;
            currentBlock = currentBlock.prev;
            return res;
        }
    }

    /**
     * Set the pointer between (n-1)-th and n-th elements of list.
     * Returns the zero-based index of element, before which the pointer is
     * actually set (for instance, the size of list, if n >= size()).
     */
    int setPointer(int n) {
	int s = n - pointerPos;
	if (s < 0) {
	    while (!inBeg() && s != 0) {
		moveBack(); s++;
	    }
	} else if (s > 0) {
	    while (!inEnd() && s != 0) {
		moveForward(); s--;
	    }
	}
        return pointerPos;
    }

    int getPointerPosition() { return pointerPos; }

    public Enumeration elements() {
        return new ListEnumeration(headList.next);
    }

    public Enumeration elementsAfter() {
        return new ListEnumeration(pointer);
    }

    public Enumeration elementsBefore() {
        return new ListReverseEnumeration(pointer.prev);
    }
}
